@@ -113,7 +113,7 @@ int do_info(nvme_mi_ep_t ep)
 	printf(" smart warnings:    0x%x\n", ss_health.sw);
 	printf(" composite temp:    %d\n", ss_health.ctemp);
 	printf(" drive life used:   %d%%\n", ss_health.pdlu);
-	printf(" controller status: 0x%04x\n", le16_to_cpu(ss_health.pdlu));
+	printf(" controller status: 0x%04x\n", le16_to_cpu(ss_health.ccs));
 
 	return 0;
 }
@@ -201,7 +201,7 @@ int do_identify(nvme_mi_ep_t ep, int argc, char **argv)
 
 	partial = argc > 2 && !strcmp(argv[2], "--partial");
 
-	ctrl = nvme_mi_init_ctrl(ep, tmp);
+	ctrl = nvme_mi_init_ctrl(ep, ctrl_id);
 	if (!ctrl) {
 		warn("can't create controller");
 		return -1;
@@ -211,7 +211,7 @@ int do_identify(nvme_mi_ep_t ep, int argc, char **argv)
 	id_args.args_size = sizeof(id_args);
 	id_args.cns = NVME_IDENTIFY_CNS_CTRL;
 	id_args.nsid = NVME_NSID_NONE;
-	id_args.cntid = ctrl_id;
+	id_args.cntid = 0;
 	id_args.csi = NVME_CSI_NVM;
 
 	/* for this example code, we can either do a full or partial identify;
@@ -311,7 +311,7 @@ int do_get_log_page(nvme_mi_ep_t ep, int argc, char **argv)
 		return -1;
 	}
 
-	rc = nvme_mi_admin_get_log_page(ctrl, &args);
+	rc = nvme_mi_admin_get_log(ctrl, &args);
 	if (rc) {
 		warn("can't perform Get Log page command");
 		return -1;
@@ -347,14 +347,14 @@ int do_admin_raw(nvme_mi_ep_t ep, int argc, char **argv)
 	}
 
 	tmp = atoi(argv[1]);
-	if (tmp < 0 || tmp > 0xffff) {
+	if (tmp > 0xffff) {
 		fprintf(stderr, "invalid controller ID\n");
 		return -1;
 	}
 	ctrl_id = tmp & 0xffff;
 
 	tmp = atoi(argv[2]);
-	if (tmp < 0 || tmp > 0xff) {
+	if (tmp > 0xff) {
 		fprintf(stderr, "invalid opcode\n");
 		return -1;
 	}
@@ -461,7 +461,7 @@ int do_security_info(nvme_mi_ep_t ep, int argc, char **argv)
 	}
 
 	tmp = atoi(argv[1]);
-	if (tmp < 0 || tmp > 0xffff) {
+	if (tmp > 0xffff) {
 		fprintf(stderr, "invalid controller ID\n");
 		return -1;
 	}
@@ -661,6 +661,12 @@ static int do_action_endpoint(enum action action, nvme_mi_ep_t ep, int argc, cha
 	case ACTION_CONFIG_SET:
 		rc = do_config_set(ep, argc, argv);
 		break;
+	default:
+		/* This shouldn't be possible, as we should be covering all
+		 * of the enum action options above. Hoever, keep the compilers
+		 * happy and fail gracefully. */
+		fprintf(stderr, "invalid action %d?\n", action);
+		rc = -1;
 	}
 	return rc;
 }
@@ -671,8 +677,8 @@ int main(int argc, char **argv)
 	nvme_root_t root;
 	nvme_mi_ep_t ep;
 	bool dbus = false, usage = true;
-	uint8_t eid;
-	int rc = 0, net;
+	uint8_t eid = 0;
+	int rc = 0, net = 0;
 
 	if (argc >= 2 && strcmp(argv[1], "dbus") == 0) {
 		usage = false;
