@@ -13,15 +13,18 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <inttypes.h>
 
 #include <sys/ioctl.h>
 #include <sys/stat.h>
+#include <sys/time.h>
 
 #include <ccan/build_assert/build_assert.h>
 #include <ccan/endian/endian.h>
 
 #include "ioctl.h"
 #include "util.h"
+#include "log.h"
 
 static int nvme_verify_chr(int fd)
 {
@@ -75,9 +78,10 @@ int nvme_get_nsid(int fd, __u32 *nsid)
 	return -1 * (errno != 0);
 }
 
-static int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
-				  struct nvme_passthru_cmd64 *cmd,
-				  __u64 *result)
+__attribute__((weak))
+int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
+			   struct nvme_passthru_cmd64 *cmd,
+			   __u64 *result)
 {
 	int err = ioctl(fd, ioctl_cmd, cmd);
 
@@ -86,8 +90,9 @@ static int nvme_submit_passthru64(int fd, unsigned long ioctl_cmd,
 	return err;
 }
 
-static int nvme_submit_passthru(int fd, unsigned long ioctl_cmd,
-				struct nvme_passthru_cmd *cmd, __u32 *result)
+__attribute__((weak))
+int nvme_submit_passthru(int fd, unsigned long ioctl_cmd,
+			 struct nvme_passthru_cmd *cmd, __u32 *result)
 {
 	int err = ioctl(fd, ioctl_cmd, cmd);
 
@@ -193,125 +198,6 @@ int nvme_admin_passthru(int fd, __u8 opcode, __u8 flags, __u16 rsvd,
 			     metadata, timeout_ms, result);
 }
 
-enum nvme_cmd_dword_fields {
-	NVME_DEVICE_SELF_TEST_CDW10_STC_SHIFT			= 0,
-	NVME_DEVICE_SELF_TEST_CDW10_STC_MASK			= 0xf,
-	NVME_DIRECTIVE_CDW11_DOPER_SHIFT			= 0,
-	NVME_DIRECTIVE_CDW11_DTYPE_SHIFT			= 8,
-	NVME_DIRECTIVE_CDW11_DPSEC_SHIFT			= 16,
-	NVME_DIRECTIVE_CDW11_DOPER_MASK				= 0xff,
-	NVME_DIRECTIVE_CDW11_DTYPE_MASK				= 0xff,
-	NVME_DIRECTIVE_CDW11_DPSEC_MASK				= 0xffff,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_ENDIR_SHIFT		= 0,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_DTYPE_SHIFT		= 1,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_ENDIR_MASK		= 0x1,
-	NVME_DIRECTIVE_SEND_IDENTIFY_CDW12_DTYPE_MASK		= 0x1,
-	NVME_FW_COMMIT_CDW10_FS_SHIFT				= 0,
-	NVME_FW_COMMIT_CDW10_CA_SHIFT				= 3,
-	NVME_FW_COMMIT_CDW10_BPID_SHIFT				= 31,
-	NVME_FW_COMMIT_CDW10_FS_MASK				= 0x7,
-	NVME_FW_COMMIT_CDW10_CA_MASK				= 0x7,
-	NVME_FW_COMMIT_CDW10_BPID_MASK				= 0x1,
-	NVME_GET_FEATURES_CDW10_SEL_SHIFT			= 8,
-	NVME_GET_FEATURES_CDW10_SEL_MASK			= 0x7,
-	NVME_SET_FEATURES_CDW10_SAVE_SHIFT			= 31,
-	NVME_SET_FEATURES_CDW10_SAVE_MASK			= 0x1,
-	NVME_FEATURES_CDW10_FID_SHIFT				= 0,
-	NVME_FEATURES_CDW14_UUID_SHIFT				= 0,
-	NVME_FEATURES_CDW10_FID_MASK				= 0xff,
-	NVME_FEATURES_CDW14_UUID_MASK				= 0x7f,
-	NVME_LOG_CDW10_LID_SHIFT				= 0,
-	NVME_LOG_CDW10_LSP_SHIFT				= 8,
-	NVME_LOG_CDW10_RAE_SHIFT				= 15,
-	NVME_LOG_CDW10_NUMDL_SHIFT				= 16,
-	NVME_LOG_CDW11_NUMDU_SHIFT				= 0,
-	NVME_LOG_CDW11_LSI_SHIFT				= 16,
-	NVME_LOG_CDW14_UUID_SHIFT				= 0,
-	NVME_LOG_CDW14_CSI_SHIFT				= 24,
-	NVME_LOG_CDW14_OT_SHIFT					= 23,
-	NVME_LOG_CDW10_LID_MASK					= 0xff,
-	NVME_LOG_CDW10_LSP_MASK					= 0x7f,
-	NVME_LOG_CDW10_RAE_MASK					= 0x1,
-	NVME_LOG_CDW10_NUMDL_MASK				= 0xffff,
-	NVME_LOG_CDW11_NUMDU_MASK				= 0xffff,
-	NVME_LOG_CDW11_LSI_MASK					= 0xffff,
-	NVME_LOG_CDW14_UUID_MASK				= 0x7f,
-	NVME_LOG_CDW14_CSI_MASK					= 0xff,
-	NVME_LOG_CDW14_OT_MASK					= 0x1,
-	NVME_IDENTIFY_CDW10_CNS_SHIFT				= 0,
-	NVME_IDENTIFY_CDW10_CNTID_SHIFT				= 16,
-	NVME_IDENTIFY_CDW11_CNSSPECID_SHIFT			= 0,
-	NVME_IDENTIFY_CDW14_UUID_SHIFT				= 0,
-	NVME_IDENTIFY_CDW11_CSI_SHIFT				= 24,
-	NVME_IDENTIFY_CDW10_CNS_MASK				= 0xff,
-	NVME_IDENTIFY_CDW10_CNTID_MASK				= 0xffff,
-	NVME_IDENTIFY_CDW11_CNSSPECID_MASK			= 0xffff,
-	NVME_IDENTIFY_CDW14_UUID_MASK				= 0x7f,
-	NVME_IDENTIFY_CDW11_CSI_MASK				= 0xff,
-	NVME_NAMESPACE_ATTACH_CDW10_SEL_SHIFT			= 0,
-	NVME_NAMESPACE_ATTACH_CDW10_SEL_MASK			= 0xf,
-	NVME_NAMESPACE_MGMT_CDW10_SEL_SHIFT			= 0,
-	NVME_NAMESPACE_MGMT_CDW10_SEL_MASK			= 0xf,
-	NVME_NAMESPACE_MGMT_CDW11_CSI_SHIFT			= 24,
-	NVME_NAMESPACE_MGMT_CDW11_CSI_MASK			= 0xff,
-	NVME_VIRT_MGMT_CDW10_ACT_SHIFT				= 0,
-	NVME_VIRT_MGMT_CDW10_RT_SHIFT				= 8,
-	NVME_VIRT_MGMT_CDW10_CNTLID_SHIFT			= 16,
-	NVME_VIRT_MGMT_CDW11_NR_SHIFT				= 0,
-	NVME_VIRT_MGMT_CDW10_ACT_MASK				= 0xf,
-	NVME_VIRT_MGMT_CDW10_RT_MASK				= 0x7,
-	NVME_VIRT_MGMT_CDW10_CNTLID_MASK			= 0xffff,
-	NVME_VIRT_MGMT_CDW11_NR_MASK				= 0xffff,
-	NVME_FORMAT_CDW10_LBAF_SHIFT				= 0,
-	NVME_FORMAT_CDW10_MSET_SHIFT				= 4,
-	NVME_FORMAT_CDW10_PI_SHIFT				= 5,
-	NVME_FORMAT_CDW10_PIL_SHIFT				= 8,
-	NVME_FORMAT_CDW10_SES_SHIFT				= 9,
-	NVME_FORMAT_CDW10_LBAFU_SHIFT				= 12,
-	NVME_FORMAT_CDW10_LBAF_MASK				= 0xf,
-	NVME_FORMAT_CDW10_MSET_MASK				= 0x1,
-	NVME_FORMAT_CDW10_PI_MASK				= 0x7,
-	NVME_FORMAT_CDW10_PIL_MASK				= 0x1,
-	NVME_FORMAT_CDW10_SES_MASK				= 0x7,
-	NVME_FORMAT_CDW10_LBAFU_MASK				= 0x3,
-	NVME_SANITIZE_CDW10_SANACT_SHIFT			= 0,
-	NVME_SANITIZE_CDW10_AUSE_SHIFT				= 3,
-	NVME_SANITIZE_CDW10_OWPASS_SHIFT			= 4,
-	NVME_SANITIZE_CDW10_OIPBP_SHIFT				= 8,
-	NVME_SANITIZE_CDW10_NODAS_SHIFT				= 9,
-	NVME_SANITIZE_CDW10_SANACT_MASK				= 0x7,
-	NVME_SANITIZE_CDW10_AUSE_MASK				= 0x1,
-	NVME_SANITIZE_CDW10_OWPASS_MASK				= 0xf,
-	NVME_SANITIZE_CDW10_OIPBP_MASK				= 0x1,
-	NVME_SANITIZE_CDW10_NODAS_MASK				= 0x1,
-	NVME_SECURITY_NSSF_SHIFT				= 0,
-	NVME_SECURITY_SPSP0_SHIFT				= 8,
-	NVME_SECURITY_SPSP1_SHIFT				= 16,
-	NVME_SECURITY_SECP_SHIFT				= 24,
-	NVME_SECURITY_NSSF_MASK					= 0xff,
-	NVME_SECURITY_SPSP0_MASK				= 0xff,
-	NVME_SECURITY_SPSP1_MASK				= 0xff,
-	NVME_SECURITY_SECP_MASK					= 0xffff,
-	NVME_GET_LBA_STATUS_CDW13_RL_SHIFT			= 0,
-	NVME_GET_LBA_STATUS_CDW13_ATYPE_SHIFT			= 24,
-	NVME_GET_LBA_STATUS_CDW13_RL_MASK			= 0xffff,
-	NVME_GET_LBA_STATUS_CDW13_ATYPE_MASK			= 0xff,
-	NVME_ZNS_MGMT_SEND_ZSASO_SHIFT				= 9,
-	NVME_ZNS_MGMT_SEND_ZSASO_MASK				= 0x1,
-	NVME_ZNS_MGMT_SEND_SEL_SHIFT				= 8,
-	NVME_ZNS_MGMT_SEND_SEL_MASK				= 0x1,
-	NVME_ZNS_MGMT_SEND_ZSA_SHIFT				= 0,
-	NVME_ZNS_MGMT_SEND_ZSA_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRA_SHIFT				= 0,
-	NVME_ZNS_MGMT_RECV_ZRA_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRASF_SHIFT				= 8,
-	NVME_ZNS_MGMT_RECV_ZRASF_MASK				= 0xff,
-	NVME_ZNS_MGMT_RECV_ZRAS_FEAT_SHIFT			= 16,
-	NVME_ZNS_MGMT_RECV_ZRAS_FEAT_MASK			= 0x1,
-	NVME_DIM_TAS_SHIFT					= 0,
-	NVME_DIM_TAS_MASK					= 0xF,
-};
-
 enum features {
 	NVME_FEATURES_ARBITRATION_BURST_SHIFT			= 0,
 	NVME_FEATURES_ARBITRATION_LPW_SHIFT			= 8,
@@ -434,7 +320,7 @@ int nvme_get_log_page(int fd, __u32 xfer_len, struct nvme_get_log_args *args)
 {
 	__u64 offset = 0, xfer, data_len = args->len;
 	__u64 start = args->lpo;
-	bool retain = true;
+	bool retain = args->rae;
 	void *ptr = args->log;
 	int ret;
 
@@ -454,13 +340,10 @@ int nvme_get_log_page(int fd, __u32 xfer_len, struct nvme_get_log_args *args)
 		 * last portion of this log page so the data remains latched
 		 * during the fetch sequence.
 		 */
-		if (offset + xfer == data_len)
-			retain = args->rae;
-
 		args->lpo = start + offset;
 		args->len = xfer;
 		args->log = ptr;
-		args->rae = retain;
+		args->rae = offset + xfer < data_len || retain;
 		ret = nvme_get_log(args);
 		if (ret)
 			return ret;
@@ -535,16 +418,18 @@ int nvme_set_features_power_mgmt(int fd, __u8 ps, __u8 wh, bool save,
 				 __u32 *result)
 {
 	__u32 value = NVME_SET(ps, FEAT_PWRMGMT_PS) |
-			NVME_SET(wh, FEAT_PWRMGMT_PS);
+			NVME_SET(wh, FEAT_PWRMGMT_WH);
 
 	return __nvme_set_features(fd, NVME_FEAT_FID_POWER_MGMT, value, save,
 				   result);
 }
 
-int nvme_set_features_lba_range(int fd, __u32 nsid, __u32 nr_ranges, bool save,
+int nvme_set_features_lba_range(int fd, __u32 nsid, __u8 nr_ranges, bool save,
 				struct nvme_lba_range_type *data, __u32 *result)
 {
-	return -1;
+	return nvme_set_features_data(
+		fd, NVME_FEAT_FID_LBA_RANGE, nsid, nr_ranges - 1, save,
+		sizeof(*data), data, result);
 }
 
 int nvme_set_features_temp_thresh(int fd, __u16 tmpth, __u8 tmpsel,
@@ -565,8 +450,8 @@ int nvme_set_features_err_recovery(int fd, __u32 nsid, __u16 tler, bool dulbe,
 	__u32 value = NVME_SET(tler, FEAT_ERROR_RECOVERY_TLER) |
 			NVME_SET(!!dulbe, FEAT_ERROR_RECOVERY_DULBE);
 
-	return __nvme_set_features(fd, NVME_FEAT_FID_ERR_RECOVERY, value, save,
-				   result);
+	return nvme_set_features_simple(
+		fd, NVME_FEAT_FID_ERR_RECOVERY, nsid, value, save, result);
 }
 
 int nvme_set_features_volatile_wc(int fd, bool wce, bool save, __u32 *result)
@@ -580,8 +465,8 @@ int nvme_set_features_volatile_wc(int fd, bool wce, bool save, __u32 *result)
 int nvme_set_features_irq_coalesce(int fd, __u8 thr, __u8 time, bool save,
 				   __u32 *result)
 {
-	__u32 value = NVME_SET(thr, FEAT_IRQC_TIME) |
-			NVME_SET(time, FEAT_IRQC_THR);
+	__u32 value = NVME_SET(thr, FEAT_IRQC_THR) |
+			NVME_SET(time, FEAT_IRQC_TIME);
 
 	return __nvme_set_features(fd, NVME_FEAT_FID_IRQ_COALESCE, value, save,
 				   result);
@@ -615,19 +500,31 @@ int nvme_set_features_async_event(int fd, __u32 events,
 int nvme_set_features_auto_pst(int fd, bool apste, bool save,
 			       struct nvme_feat_auto_pst *apst, __u32 *result)
 {
-	__u32 value = NVME_SET(!!apste, FEAT_APST_APSTE);
+	struct nvme_set_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_AUTO_PST,
+		.nsid = NVME_NSID_NONE,
+		.cdw11 = NVME_SET(!!apste, FEAT_APST_APSTE),
+		.save = save,
+		.uuidx = NVME_UUID_NONE,
+		.data = apst,
+		.data_len = sizeof(*apst),
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
 
-	return __nvme_set_features(fd, NVME_FEAT_FID_AUTO_PST, value, save,
-				   result);
+	return nvme_set_features(&args);
 }
 
 int nvme_set_features_timestamp(int fd, bool save, __u64 timestamp)
 {
 	__le64 t = cpu_to_le64(timestamp);
-	struct nvme_timestamp ts;
+	struct nvme_timestamp ts = {};
 	struct nvme_set_features_args args = {
 		.args_size = sizeof(args),
 		.fd = fd,
+		.fid = NVME_FEAT_FID_TIMESTAMP,
 		.nsid = NVME_NSID_NONE,
 		.cdw11 = 0,
 		.cdw12 = 0,
@@ -697,8 +594,8 @@ int nvme_set_features_plm_config(int fd, bool plm, __u16 nvmsetid, bool save,
 		.save = save,
 		.uuidx = NVME_UUID_NONE,
 		.cdw15 = 0,
-		.data_len = 0,
-		.data = NULL,
+		.data_len = sizeof(*data),
+		.data = data,
 		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
 		.result = result,
 	};
@@ -749,7 +646,7 @@ int nvme_set_features_host_behavior(int fd, bool save,
 		.nsid = NVME_NSID_NONE,
 		.cdw11 = 0,
 		.cdw12 = 0,
-		.save = save,
+		.save = false,
 		.uuidx = NVME_UUID_NONE,
 		.cdw15 = 0,
 		.data_len = sizeof(*data),
@@ -783,7 +680,7 @@ int nvme_set_features_sw_progress(int fd, __u8 pbslc, bool save,
 				   result);
 }
 
-int nvme_set_features_host_id(int fd, bool save, bool exhid, __u8 *hostid)
+int nvme_set_features_host_id(int fd, bool exhid, bool save, __u8 *hostid)
 {
 	__u32 len = exhid ? 16 : 8;
 	__u32 value = !!exhid;
@@ -812,20 +709,42 @@ int nvme_set_features_resv_mask(int fd, __u32 mask, bool save, __u32 *result)
 				   result);
 }
 
+int nvme_set_features_resv_mask2(int fd, __u32 nsid, __u32 mask, bool save,
+				 __u32 *result)
+{
+	return nvme_set_features_simple(
+		fd, NVME_FEAT_FID_RESV_MASK, nsid, mask, save, result);
+}
+
 int nvme_set_features_resv_persist(int fd, bool ptpl, bool save, __u32 *result)
 {
 	return __nvme_set_features(fd, NVME_FEAT_FID_RESV_PERSIST, !!ptpl, save,
 				   result);
 }
 
+int nvme_set_features_resv_persist2(int fd, __u32 nsid, bool ptpl, bool save,
+				    __u32 *result)
+{
+	return nvme_set_features_simple(
+		fd, NVME_FEAT_FID_RESV_PERSIST, nsid, !!ptpl, save, result);
+}
+
 int nvme_set_features_write_protect(int fd, enum nvme_feat_nswpcfg_state state,
 				    bool save, __u32 *result)
 {
 	return __nvme_set_features(fd, NVME_FEAT_FID_WRITE_PROTECT, state,
-				   save, result);
+				   false, result);
 }
 
-int nvme_set_features_iocs_profile(int fd, __u8 iocsi, bool save)
+int nvme_set_features_write_protect2(int fd, __u32 nsid,
+				     enum nvme_feat_nswpcfg_state state,
+				     bool save, __u32 *result)
+{
+	return nvme_set_features_simple(
+		fd, NVME_FEAT_FID_WRITE_PROTECT, nsid, state, false, result);
+}
+
+int nvme_set_features_iocs_profile(int fd, __u16 iocsi, bool save)
 {
 	__u32 value = NVME_SET(iocsi, FEAT_IOCSP_IOCSCI);
 
@@ -901,8 +820,28 @@ int nvme_get_features_lba_range(int fd, enum nvme_get_features_sel sel,
 		.sel = sel,
 		.cdw11 = 0,
 		.uuidx = NVME_UUID_NONE,
-		.data_len = 0,
-		.data = NULL,
+		.data_len = sizeof(*data),
+		.data = data,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
+int nvme_get_features_lba_range2(int fd, enum nvme_get_features_sel sel,
+				 __u32 nsid, struct nvme_lba_range_type *data,
+				 __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_LBA_RANGE,
+		.nsid = nsid,
+		.sel = sel,
+		.uuidx = NVME_UUID_NONE,
+		.data = data,
+		.data_len = sizeof(*data),
 		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
 		.result = result,
 	};
@@ -923,6 +862,24 @@ int nvme_get_features_err_recovery(int fd, enum nvme_get_features_sel sel,
 				   result);
 }
 
+int nvme_get_features_err_recovery2(int fd, enum nvme_get_features_sel sel,
+				    __u32 nsid, __u32 *result)
+{
+
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_ERR_RECOVERY,
+		.nsid = nsid,
+		.sel = sel,
+		.uuidx = NVME_UUID_NONE,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
 int nvme_get_features_volatile_wc(int fd, enum nvme_get_features_sel sel,
 				  __u32 *result)
 {
@@ -948,7 +905,7 @@ int nvme_get_features_irq_config(int fd, enum nvme_get_features_sel sel,
 	struct nvme_get_features_args args = {
 		.args_size = sizeof(args),
 		.fd = fd,
-		.fid = NVME_FEAT_FID_LBA_RANGE,
+		.fid = NVME_FEAT_FID_IRQ_CONFIG,
 		.nsid = NVME_NSID_NONE,
 		.sel = sel,
 		.cdw11 = iv,
@@ -981,13 +938,13 @@ int nvme_get_features_auto_pst(int fd, enum nvme_get_features_sel sel,
 	struct nvme_get_features_args args = {
 		.args_size = sizeof(args),
 		.fd = fd,
-		.fid = NVME_FEAT_FID_LBA_RANGE,
+		.fid = NVME_FEAT_FID_AUTO_PST,
 		.nsid = NVME_NSID_NONE,
 		.sel = sel,
 		.cdw11 = 0,
 		.uuidx = NVME_UUID_NONE,
-		.data_len = 0,
-		.data = NULL,
+		.data_len = sizeof(*apst),
+		.data = apst,
 		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
 		.result = result,
 	};
@@ -1001,6 +958,26 @@ int nvme_get_features_host_mem_buf(int fd, enum nvme_get_features_sel sel,
 	return __nvme_get_features(fd, NVME_FEAT_FID_HOST_MEM_BUF, sel, result);
 }
 
+int nvme_get_features_host_mem_buf2(int fd, enum nvme_get_features_sel sel,
+				    struct nvme_host_mem_buf_attrs *attrs,
+				    __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_HOST_MEM_BUF,
+		.nsid = NVME_NSID_NONE,
+		.sel = sel,
+		.uuidx = NVME_UUID_NONE,
+		.data = attrs,
+		.data_len = sizeof(*attrs),
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
 int nvme_get_features_timestamp(int fd, enum nvme_get_features_sel sel,
 				struct nvme_timestamp *ts)
 {
@@ -1053,8 +1030,8 @@ int nvme_get_features_plm_config(int fd, enum nvme_get_features_sel sel,
 		.sel = sel,
 		.cdw11 = nvmsetid,
 		.uuidx = NVME_UUID_NONE,
-		.data_len = 0,
-		.data = NULL,
+		.data_len = sizeof(*data),
+		.data = data,
 		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
 		.result = result,
 	};
@@ -1101,8 +1078,8 @@ int nvme_get_features_host_behavior(int fd, enum nvme_get_features_sel sel,
 		.sel = sel,
 		.cdw11 = 0,
 		.uuidx = NVME_UUID_NONE,
-		.data_len = 0,
-		.data = NULL,
+		.data_len = sizeof(*data),
+		.data = data,
 		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
 		.result = result,
 	};
@@ -1125,7 +1102,7 @@ int nvme_get_features_endurance_event_cfg(int fd, enum nvme_get_features_sel sel
 		.fid = NVME_FEAT_FID_ENDURANCE_EVT_CFG,
 		.nsid = NVME_NSID_NONE,
 		.sel = sel,
-		.cdw11 = 0,
+		.cdw11 = endgid,
 		.uuidx = NVME_UUID_NONE,
 		.data_len = 0,
 		.data = NULL,
@@ -1168,12 +1145,46 @@ int nvme_get_features_resv_mask(int fd, enum nvme_get_features_sel sel,
 	return __nvme_get_features(fd, NVME_FEAT_FID_RESV_MASK, sel, result);
 }
 
+int nvme_get_features_resv_mask2(int fd, enum nvme_get_features_sel sel,
+				 __u32 nsid, __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_RESV_MASK,
+		.nsid = nsid,
+		.sel = sel,
+		.uuidx = NVME_UUID_NONE,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
 int nvme_get_features_resv_persist(int fd, enum nvme_get_features_sel sel,
 				   __u32 *result)
 {
 	return __nvme_get_features(fd, NVME_FEAT_FID_RESV_PERSIST, sel, result);
 }
 
+int nvme_get_features_resv_persist2(int fd, enum nvme_get_features_sel sel,
+				    __u32 nsid, __u32 *result)
+{
+	struct nvme_get_features_args args = {
+		.args_size = sizeof(args),
+		.fd = fd,
+		.fid = NVME_FEAT_FID_RESV_PERSIST,
+		.nsid = nsid,
+		.sel = sel,
+		.uuidx = NVME_UUID_NONE,
+		.timeout = NVME_DEFAULT_IOCTL_TIMEOUT,
+		.result = result,
+	};
+
+	return nvme_get_features(&args);
+}
+
 int nvme_get_features_write_protect(int fd, __u32 nsid,
 				    enum nvme_get_features_sel sel,
 				    __u32 *result)
@@ -1235,9 +1246,15 @@ int nvme_format_nvm(struct nvme_format_nvm_args *args)
 
 int nvme_ns_mgmt(struct nvme_ns_mgmt_args *args)
 {
+	const size_t size_v1 = sizeof_args(struct nvme_ns_mgmt_args, csi, __u64);
+	const size_t size_v2 = sizeof_args(struct nvme_ns_mgmt_args, data, __u64);
 	__u32 cdw10    = NVME_SET(args->sel, NAMESPACE_MGMT_CDW10_SEL);
 	__u32 cdw11    = NVME_SET(args->csi, NAMESPACE_MGMT_CDW11_CSI);
-	__u32 data_len = args->ns ? sizeof(*args->ns) : 0;
+
+	if (args->args_size < size_v1 || args->args_size > size_v2) {
+		errno = EINVAL;
+		return -1;
+	}
 
 	struct nvme_passthru_cmd cmd = {
 		.nsid	    = args->nsid,
@@ -1245,13 +1262,19 @@ int nvme_ns_mgmt(struct nvme_ns_mgmt_args *args)
 		.cdw10	    = cdw10,
 		.cdw11	    = cdw11,
 		.timeout_ms = args->timeout,
-		.data_len   = data_len,
-		.addr	    = (__u64)(uintptr_t)args->ns,
 	};
 
-	if (args->args_size < sizeof(*args)) {
-		errno = EINVAL;
-		return -1;
+	if (args->args_size == size_v2) {
+		if (args->data) {
+			cmd.data_len = sizeof(*args->data);
+			cmd.addr = (__u64)(uintptr_t)args->data;
+		}
+	}
+	else {
+		if (args->ns) {
+			cmd.data_len = sizeof(*args->ns);
+			cmd.addr = (__u64)(uintptr_t)args->ns;
+		}
 	}
 	return nvme_submit_admin_passthru(args->fd, &cmd, args->result);
 }
@@ -1378,6 +1401,7 @@ int nvme_get_lba_status(struct nvme_get_lba_status_args *args)
 		.opcode =  nvme_admin_get_lba_status,
 		.nsid = args->nsid,
 		.addr = (__u64)(uintptr_t)args->lbas,
+		.data_len = (args->mndw + 1) << 2,
 		.cdw10 = cdw10,
 		.cdw11 = cdw11,
 		.cdw12 = cdw12,
@@ -1646,33 +1670,35 @@ static int nvme_set_var_size_tags(__u32 *cmd_dw2, __u32 *cmd_dw3, __u32 *cmd_dw1
 		__u8 pif, __u8 sts, __u64 reftag, __u64 storage_tag)
 {
 	__u32 cdw2 = 0, cdw3 = 0, cdw14;
+	beint64_t be_reftag = cpu_to_be64(reftag);
+	beint64_t be_storage_tag = cpu_to_be64(storage_tag);
 
 	switch (pif) {
 	/* 16b Protection Information */
 	case 0:
-		cdw14 = reftag & 0xffffffff;
-		cdw14 |= ((storage_tag << (32 - sts)) & 0xffffffff);
+		cdw14 = be_reftag & 0xffffffff;
+		cdw14 |= ((be_storage_tag << (32 - sts)) & 0xffffffff);
 		break;
 	/* 32b Protection Information */
 	case 1:
-		cdw14 = reftag & 0xffffffff;
-		cdw3 = reftag >> 32;
-		cdw14 |= ((storage_tag << (80 - sts)) & 0xffff0000);
+		cdw14 = be_reftag & 0xffffffff;
+		cdw3 = be_reftag >> 32;
+		cdw14 |= ((be_storage_tag << (80 - sts)) & 0xffff0000);
 		if (sts >= 48)
-			cdw3 |= ((storage_tag >> (sts - 48)) & 0xffffffff);
+			cdw3 |= ((be_storage_tag >> (sts - 48)) & 0xffffffff);
 		else
-			cdw3 |= ((storage_tag << (48 - sts)) & 0xffffffff);
-		cdw2 = (storage_tag >> (sts - 16)) & 0xffff;
+			cdw3 |= ((be_storage_tag << (48 - sts)) & 0xffffffff);
+		cdw2 = (be_storage_tag >> (sts - 16)) & 0xffff;
 		break;
 	/* 64b Protection Information */
 	case 2:
-		cdw14 = reftag & 0xffffffff;
-		cdw3 = (reftag >> 32) & 0xffff;
-		cdw14 |= ((storage_tag << (48 - sts)) & 0xffffffff);
+		cdw14 = be_reftag & 0xffffffff;
+		cdw3 = (be_reftag >> 32) & 0xffff;
+		cdw14 |= ((be_storage_tag << (48 - sts)) & 0xffffffff);
 		if (sts >= 16)
-			cdw3 |= ((storage_tag >> (sts - 16)) & 0xffff);
+			cdw3 |= ((be_storage_tag >> (sts - 16)) & 0xffff);
 		else
-			cdw3 |= ((storage_tag << (16 - sts)) & 0xffff);
+			cdw3 |= ((be_storage_tag << (16 - sts)) & 0xffff);
 		break;
 	default:
 		perror("Unsupported Protection Information Format");
@@ -1784,6 +1810,10 @@ int nvme_copy(struct nvme_copy_args *args)
 
 	if (args->format == 1)
 		data_len = args->nr * sizeof(struct nvme_copy_range_f1);
+	else if (args->format == 2)
+		data_len = args->nr * sizeof(struct nvme_copy_range_f2);
+	else if (args->format == 3)
+		data_len = args->nr * sizeof(struct nvme_copy_range_f3);
 	else
 		data_len = args->nr * sizeof(struct nvme_copy_range);
 
@@ -1901,7 +1931,7 @@ int nvme_resv_report(struct nvme_resv_report_args *args)
 
 int nvme_io_mgmt_recv(struct nvme_io_mgmt_recv_args *args)
 {
-	__u32 cdw10 = (args->mo & 0xf) | (args->mos & 0xff << 16);
+	__u32 cdw10 = args->mo | (args->mos << 16);
 	__u32 cdw11 = (args->data_len >> 2) - 1;
 
 	struct nvme_passthru_cmd cmd = {
@@ -1924,7 +1954,7 @@ int nvme_io_mgmt_recv(struct nvme_io_mgmt_recv_args *args)
 
 int nvme_io_mgmt_send(struct nvme_io_mgmt_send_args *args)
 {
-	__u32 cdw10 = (args->mo & 0xf) | ((args->mos & 0xff) << 16);
+	__u32 cdw10 = args->mo | (args->mos << 16);
 
 	struct nvme_passthru_cmd cmd = {
 		.opcode		= nvme_cmd_io_mgmt_send,
